#!/usr/bin/perl -w

package RPM::Source::Tools::SourceAnalyzer;

use Cwd;
use Carp;
use strict;
use warnings;

use File::Path qw(make_path remove_tree);
use File::Temp qw/tempfile tempdir/;
use File::Basename;
use Text::ParseWords;
use DistroMap 0.11;

our $VERSION=0.005;

# TODO: proper read of m4 args in [...]

our $opt_hint='';
our ($opt_resolve_pkgconfig,$opt_allow_unknown_pkgconfig,$opt_analyze_subdir_autoconf);
our $opt_skip_implicit_pkgconfig=1;

our @LONGOPT=(
    'sourcedep-hint=s'   => \$opt_hint,
    'sourcedep-pkgconfig-resolve!' => \$opt_resolve_pkgconfig,
    'sourcedep-pkgconfig-allow-unknown!' => \$opt_allow_unknown_pkgconfig,
    'sourcedep-pkgconfig-skip-implicit!' => \$opt_skip_implicit_pkgconfig,
    'sourcedep-autoconf-search-subdirs!' => \$opt_analyze_subdir_autoconf,
    );

sub longopt_usage {
    print "SourceAnalyzer Options:
    General Options.
	--sourcedep-hint <hint> - pass a hint to SourceAnalyzer core.
	  current hints are: octave convert
	--sourcedep-autoconf-search-subdirs	search subdirs for extra configure.ac/in files
	--no-sourcedep-autoconf-search-subdirs  negate the option (default)
    Options for special dependencies management.
	Special dependencies are dependencies like type(name) and are often stricter than dependencies on package names. Generation of such dependencies is ruled by options --sourcedep-<type>-*.
	--sourcedep-<type>-resolve		replace all special dependencies with package names.
	--sourcedep-<type>-skip-implicit	drop special dependencies whose packages are mentioned in spec.
	--sourcedep-<type>-allow-unknown	keep all special dependencies, even not found in the repository.

	--sourcedep-pkgconfig-resolve		use package names instead of pkgconfig(name)
	--sourcedep-pkgconfig-skip-implicit	skip pkgconfig(name) dependency if its package is present (default).
	--sourcedep-pkgconfig-allow-unknown	add pkgconfig(name) dependency even if its package is not found.
	--no-sourcedep-pkgconfig-allow-unknown	negate the option (default)
	--no-sourcedep-pkgconfig-resolve	negate the option (default)
	--no-sourcedep-pkgconfig-skip-implicit	negate the option
";
}

my %BIN_IGNORE=map {$_=>1} qw!
gcc
lib.exe
csc.exe
!;
my %LIB_IGNORE=map {$_=>1} 
# deprecated
qw!
sasl
!, # in std lang environment
qw!
gcc
stdc++
!, #other_unix
'gen', # libgen -_string_pattern-matching_library_Solaris
qw!
getopt
gnugetopt
socket
sunmath
mlib
!, # Must die
qw!
dnsapi
iconv
inet
intl
iphlpapi
resmgr
sendfile
winmm
ws2_32
wsock32
!;
my %HEADER_IGNORE=map {$_=>1} (
'Carbon/Carbon.h',	# MAC
'CoreServices/CoreServices.h', # MAC
'QuickTime/QuickTime.h', # MAC
'fp_class.h',		# *
'ieeefp.h',		# *
'machine/endian.h',	# BSD
'machine/types.h',	# BSD
'nan.h',		# BSD
'sndio.h',		# OpenBSD,...
'sunmath.h',		# Solaris
'sys/byteswap.h',	# SPARC?
'sys/dg_sys_info.h',	# AIX, HP-UX
'sys/audioio.h',	# BSD
'sys/cdio.h',		# BSD
'sys/fbio.h',		# BSD
'sys/disk.h',		# BSD
'sys/endian.h',		# BSD
'sys/extaddr.h',	# BSD
'sys/if.h',		# BSD
'sys/int_types.h',	# BSD
'sys/mode.h',		# AIX
'sys/sockio.h',		# BSD
'sys/scsiio.h',		# BSD
'sys/videoio.h',	# BSD
'sys/systeminfo.h',	# Solaris
'sys/processor.h',	# Solaris
'sys/loadavg.h',	# Solaris
'sys/fpu.h',		# *
'windows.h',	# Must die
'winsock.h',	# Must die
'winsock2.h',	# Must die
'ws2tcpip.h',	# Must die
'basetsd.h',	# Must die
'hash_map',	# Must die
'hash_set',	# Must die
'iphlpapi.h',	# Must die
'tchar.h',	# Must die
'mmsystem.h',	# Must die
'DShow.h',	# DirectShow
'strstream',
),
# extra sys headers
qw!
android/native_window.h
cygwin/fs.h
machine/soundcard.h
sys/acpa.h
sys/apparmor.h
sys/audio.h
sys/bufmod.h
sys/devpoll.h
sys/disklabel.h
sys/dl.h
sys/dlpi_ext.h
sys/dlpi.h
sys/dlpihdr.h
sys/dvdio.h
sys/event.h
sys/extattr.h
sys/fpu.h
sys/ioccom.h
sys/mib.h
sys/mixer.h
sys/mkdev.h
sys/mntent.h
sys/ndd_var.h
sys/net/nit.h
sys/proc.h
sys/sam9407.h
sys/sched.h
sys/scsi.h
sys/scsi/scsi_types.h
sys/sdt.h
sys/stdtypes.h
sys/stropts.h
sys/telephony.h
sys/uc_access.h
sys/ucred.h
sys/utime.h
sys/videodev2.h
!;

&__add_file_to_flag('/usr/share/srpmtools/data/stdheaders.txt',\%HEADER_IGNORE);

# TODO:
my %HEADER2PKG=qw/
jni.h	java-devel
/;

sub new {
    my $class = shift;
    my $self = {
	__IS_BINARY__ => 0,
	-verbose => 0,
	-current_buildreq => {},
	-hint => $opt_hint,
	-resolve_pkgconfig => $opt_resolve_pkgconfig,
	-allow_unknown_pkgconfig => $opt_allow_unknown_pkgconfig,
	-skip_implicit_pkgconfig => $opt_skip_implicit_pkgconfig,
	-analyze_subdir_autoconf => $opt_analyze_subdir_autoconf,
	@_
    };
    bless $self, $class;
    
    my $hint=$self->{-hint};
    if ($hint and $hint eq 'octave') {
	$self->{-analyze_subdir_autoconf}=1 if not defined $self->{-analyze_subdir_autoconf};
    }
    return $self;
}

sub is_arch_dependent {
    my ($self,$flag)=@_;
    return $self->{__IS_BINARY__};
}

sub __set_binary {
    my ($self,$flag)=@_;
    $self->{__IS_BINARY__}=1;
}

sub analyze {
    my ($self,$archive_obj)=@_;
    Carp::cluck "analyze: $archive_obj should be an object of SourceArchive" unless ref $archive_obj;
    my $verbose=$self->{-verbose};
    my $buildreq=$self->{-buildreq_storage};
    my $distroinfo=$self->{-distroinfo};
    Carp::croak "-distroinfo is required" unless $self->{-distroinfo};
    Carp::croak "-buildreq_storage is required" unless $self->{-buildreq_storage};

    my ($lang_c,$lang_cxx,$lang_fortran,$lang_python);

    my $filelist=$archive_obj->filelist();
    my @configurac=grep {m!(?:^|/)configure\.(?:ac|in)$!} @$filelist;
    my @makefileam=grep {m!(?:^|/)Makefile\.am$!} @$filelist;
    my @qmakepro=grep {m!\.pr[io]$!} @$filelist;
    my $hint=$self->{-hint};
    if ($hint and $hint eq 'octave') {
	unshift @configurac, grep {m!(?:^|/)configure\.(?:base|add)$!} @$filelist;
    }
    if (not $self->{-analyze_subdir_autoconf}) {
	# temporary?; let us analyze root only
	@configurac=grep {not m!^.*/.*/configure!} @configurac;
    }

    $lang_cxx=1 if grep {/\.c(?:c|pp|xx)$/ || m!(?:^|/)CMakeLists.txt$!} @$filelist;
    $lang_python=1 if grep {m!(?:^|/)__init__.py$!} @$filelist;
    if ($hint and $hint eq 'octave') {
	$lang_c=1 if grep {/\.(?:c|h)$/} @$filelist;
	$lang_fortran=1 if grep {/\.(?:f)$/} @$filelist;
    }
    $buildreq->add('gcc-c++') if $lang_cxx;
    $buildreq->add('gcc-fortran') if $lang_python;
    $buildreq->add('python-devel') if $lang_python;
    $self->__set_binary(1) if $lang_c or $lang_cxx or $lang_fortran;

    my $curdir = getcwd;
    my @unpack_list=(@configurac,@makefileam,@qmakepro);
    if (@unpack_list) {
	# unpack 
	my $tmpdir = $archive_obj->extract_files(\@unpack_list);
	foreach my $configuracfile (@configurac) {
	    $self->__analyze_configure_ac("$tmpdir/$configuracfile",$configuracfile);
	}
	foreach my $amfile (@makefileam) {
	    $self->__analyze_makefile_am("$tmpdir/$amfile",$amfile);
	}
	unless (@configurac) {
	    foreach my $qproject (@qmakepro) {
		$self->__analyze_qmakepro("$tmpdir/$qproject",$qproject);
	    }
	}
    }
}

sub __unique_list {
    my @out;
    my %req;
    foreach my $a (@_) {
	if (ref($a)) {
	    push @out, $a;
	} else {
	    $req{$a}=1;
	}
    }
    return @out, keys %req;
}

sub __analyze_configure_ac {
    my ($self,$filename,$inputname)=@_;
    &__analyze_file_using($self,$filename,$inputname,
	     'SourceAnalyzer::Language::Autoconf',
	     -macrohook => \&_autoconf_macro_handler,
	);
}

sub __analyze_makefile_am {
    my ($self,$filename,$inputname)=@_;
    &__analyze_file_using($self,$filename,$inputname,
	     'SourceAnalyzer::Language::AutoMake',
	     -varhook => \&_automake_var_handler,
	);
}

sub __analyze_qmakepro {
    my ($self,$filename,$inputname)=@_;
    &__analyze_file_using($self,$filename,$inputname,
	     'SourceAnalyzer::Language::QMake',
	     -varhook => \&_qmake_var_handler,
	);
}

sub __analyze_file_using {
    my ($self,$filename,$inputname,$class,@args)=@_;
    my $storage=__::SourceAnalyzer::Collect->new();
    my $parser=$class->new(
	-storage => $storage,
	-verbose => $self->{-verbose},
	@args,
	);
    $parser->parse_file($filename);
    $inputname||=basename($filename);
    $self->__analyze_storage($storage,$inputname);
}


sub __analyze_storage {
    my ($self,$storage,$inputname)=@_;
   
    my $verbose=$self->{-verbose};
    my $buildreq=$self->{-buildreq_storage};
    my $distroinfo=$self->{-distroinfo};
    my $current_buildreq=$self->{-current_buildreq};

    &__uniquify_storage($storage);
    &__debug_storage($storage,$inputname.' (uniquified)') if $verbose>1;

    my @BR=@{$storage->{pkgsreq}};
    my $flag_not_found;
    &__add_br(\@BR,$current_buildreq,$distroinfo,'devel-libs',$storage->{libsreq},\$flag_not_found,\%LIB_IGNORE);
    &__add_br(\@BR,$current_buildreq,$distroinfo,'headers',$storage->{hdrsreq},\$flag_not_found,\%HEADER_IGNORE);

    if ($self->{-resolve_pkgconfig}) {
	&__add_br(\@BR,$current_buildreq,$distroinfo,'pkg-config',$storage->{pkgconfreq},\$flag_not_found);
    } else {
	foreach my $name (@{$storage->{pkgconfreq}}) {
	    my @data=&__check_dups($current_buildreq,$distroinfo,'pkg-config',$name,\$flag_not_found);
	    push @BR, "pkgconfig($name)" if (@data or $self->{-allow_unknown_pkgconfig}) and
		not ($self->{-skip_implicit_pkgconfig} and &__hash_have_any($current_buildreq,\@data));
	}
    }

  PROG:
    foreach my $name (@{$storage->{pathreq}}) {
	my @names=($name);
	@names=(@$name) if ref $name;
	# like "smth --config"
	@names=map {s/^["']//;s/["']$//;s/\s.*$//;$_} @names;
	my @data;
	my $file;
      CANDIDATE:
	foreach my $candidate (@names) {
	    $candidate=~s/\$EXEEXT$//;
	    foreach my $prefix (qw!/usr/bin/ /usr/sbin/!) {
		my $file0=$prefix.$candidate;
		my @data0=$distroinfo->translate('path',$file0);
		# for alternatives in /usr/bin
		push @data0, $distroinfo->translate('provides',$file0) unless @data0;
		next PROG if @names>1 and (&__hash_have_any($current_buildreq,\@data0) or $current_buildreq->{$file0});
		if(@data0) {
		    $file=$file0;
		    @data=@data0;
		    last CANDIDATE;
		}
	    }
	    foreach my $prefix (qw!/bin/ /sbin/!) {
		my $file0=$prefix.$candidate;
		my @data0=$distroinfo->translate('path',$file0);
		next PROG if @data0; # /bin utilities are considered as always present basesytem
	    }
	}
	if (@data>1) {
	    next PROG if &__hash_have_any($current_buildreq,\@data) or $current_buildreq->{$file};
	    warn "INFO: SourceAnalyzer: multiple provides for $file: ",join(' ',@data),"\n";
	} elsif (not @data and &__names_are_not_ignored(\%BIN_IGNORE,@names)) {
	    warn "INFO: SourceAnalyzer: can't find ".join(' ',@names)." in /usr/bin /usr/sbin.\n";
	    $flag_not_found=1;
	}
	&__push(\@BR, $file) if $file;
    }
    
    @BR = sort &__unique_list(@BR);
    print STDERR "$inputname: BR(detected):", join(' ',@BR),"\n" if $verbose >1;
    $buildreq->add_for("guessed from $inputname: ",@BR);
    
    my $lang=$storage->{lang};
    $buildreq->add('gcc-c++') if $lang->{cxx};
    $buildreq->add('gcc-fortran') if $lang->{fortran};
    $buildreq->add('python-devel') if $lang->{python};

    # TODO: 'convert' hint?
    # broken impls of AC_PROG_CC might require g++ to be installed :( see libmodelfile

    warn "SourceAnalyzer: some deps weren't found. Is DistroMap database outdated?\n" if $flag_not_found;
}

sub __add_br {
    my ($BR,$current_buildreq,$distroinfo,$db,$list,$flag_not_found_ref,$ignore)=@_;
    &__push($BR, map {&__check_dups($current_buildreq,$distroinfo,$db,$_,$flag_not_found_ref,$ignore)} @$list);
}

sub __uniquify_storage {
    my ($storage)=@_;
    foreach (qw/libsreq hdrsreq pathreq pkgconfreq/) {
	my @new_list=&__unique_list(@{$storage->{$_}});
	$storage->{$_}=\@new_list;
    }
}

sub __debug_storage {
    my ($storage,$inputname)=@_;
    &__debug_msg($inputname,'devel-libs',$storage->{libsreq});
    &__debug_msg($inputname,'headers',$storage->{hdrsreq});
    &__debug_msg($inputname,'path',$storage->{pathreq});
    &__debug_msg($inputname,'pkg-config',$storage->{pkgconfreq});
}

sub __debug_msg {
    my ($file,$br_name,$list)=@_;
    return unless @$list;
    print STDERR "$file: BR $br_name(detected):",
    join(' ',map {ref($_) eq 'ARRAY' ? '[ '.join(' ',@$_).' ]' : $_} @$list),"\n";
}

sub __check_dups {
    my ($current_buildreq,$distroinfo,$db,$name,$flag_not_found_ref,$ignore)=@_;
    my @names=($name);
    @names=@$name if ref($name) eq 'ARRAY';
    my @data;
    foreach my $candidate (@names) {
	my @data0=$distroinfo->translate($db,$candidate);
	return() if @names>1 and &__hash_have_any($current_buildreq,\@data0);
	push @data, @data0;
	last if @data;
    }
    if (@data>1) {
	return() if &__hash_have_any($current_buildreq,\@data);
	warn 'WARNING: SourceAnalyzer: multiple provides for '.join(' ',@names).': '.join(' ',@data)."\n";
    } elsif (not @data and &__names_are_not_ignored($ignore,@names)) {
	warn "INFO: SourceAnalyzer: nothing in $db provides ".join(' ',@names).".\n";
	$$flag_not_found_ref=1 if $flag_not_found_ref;
    }
    return @data;
}

sub __names_are_not_ignored {
    my ($ignore,@names)=@_;
    return 1 if not $ignore;
    foreach my $name (@names) {
	return 1 if not $ignore->{$name};
    }
    return 0;
}

sub __hash_have_any {
    my ($hash,$list)=@_;
    foreach my $key (@$list) {
	return 1 if $hash->{$key};
    }
    return 0;
}

my %LANG_FLAGS=qw/
AC_PROG_CC	c
AM_PROG_CC_STDC	c
AM_PROG_LIBTOOL	c
AC_PROG_CXX	cxx
AC_PROG_CXXCPP	cxx
AC_PROG_F77	fortran
AC_LANG_C	c
AC_LANG_CPLUSPLUS	cxx
AC_LANG_FORTRAN77	fortran
/;

my %LIB_FLAGS=qw/
AM_PATH_AALIB	aa
AM_PATH_ALSA	asound
AM_PATH_ASPELL	aspell
AC_PATH_BLUETOOTH	bluetooth
AM_PATH_CHECK	check
AC_PATH_CLN	cln
AM_PATH_CPPUNIT	cppunit
AM_PATH_DVDNAV	dvdnav
AM_PATH_ESD	esd
AM_PATH_GPG_ERROR	gpg-error
AM_PATH_GSL	gsl
AM_PATH_GTK	gtk
AM_PATH_GTK_2_0	gtk-x11-2.0
AM_PATH_GLIB	glib
AM_PATH_GLIB_2_0	glib-2.0
AM_PATH_GPGME	gpgme
AM_PATH_LIBEXPAT	expat
AM_PATH_LIBFAME	fame
AM_PATH_LIBFLAC	flac
AM_PATH_LIBGCRYPT_CCRTP	gcrypt
AM_PATH_LIBGCRYPT	gcrypt
AM_PATH_LIBPRELUDE	prelude
AM_PATH_LUA	lua
AC_PATH_MAGIC	magic
AM_PATH_OPENGL	GL
AM_PATH_OPENOBEX	openobex
AM_PATH_PHYSFS	physfs
AM_PATH_PILOT_LINK	pisync
AM_PATH_PNG	png
AM_PATH_RADIUS_ENGINE	radius-engine
AM_PATH_SC	mpqc
AM_PATH_SDL	SDL
AM_PATH_SDL_GFX	SDL_gfx
AM_PATH_SDL_IMAGE	SDL_image
AM_PATH_SDL_MIXER	SDL_mixer
AM_PATH_SDL_NET	SDL_net
AM_PATH_SDL_PANGO	SDL_pango
AM_PATH_SDL_SOUND	SDL_sound
AM_PATH_SDL_TTF	SDL_ttf
AM_PATH_TALKFILTERS	talkfilters
AC_PATH_USB	usb-1.0
AM_PATH_XML2	xml2
AC_PATH_XTRA	ICE;SM
AST_IMLIB2_SUPPORT	Imlib2;freetype;Xext
DAP_CHECK_ZLIB	z
/;

my %HEADER_FLAGS=qw!
AC_PATH_IRDA	lib/irda.h
AC_PATH_IRDA_LINUX	lib/irda.h
AC_PATH_X	X11/Xlib.h
!;

my %PKG_FLAGS=qw!
AC_PATH_KDE	kdelibs-devel
AM_PATH_PYTHON	python-devel
AC_PATH_QT_1_3	libqt3-devel
AM_PATH_RUBY	libruby-devel
!;

my %BIN_FLAGS=qw!
GTK_DOC_CHECK	gtkdocize
!;

my %IGNORE_FLAGS=map {$_=>1} qw!
AM_PATH_LISPDIR	
AC_PATH_WIN32	
AM_PATH_DIRECTX	
AM_PATH_LIBOLE2	
AM_PATH_ARTS	
AC_PATH_TOOL	
AC_PATH_GENERIC
!;

#GNOME_COMMON_INIT
## Docs are disabled for now
##GNOME_DOC_INIT
#GNOME_MAINTAINER_MODE_DEFINES
#GNOME_COMPILE_WARNINGS
#GNOME_DEBUG_CHECK

# TODO: implicit AC_TRY_LINK
#LIBS="-lutil"
#AC_TRY_LINK(, , [needutil=yes], [needutil=no])

# dep on GConf
#AM_GCONF_SOURCE_2

sub _autoconf_macro_handler {
    my ($name,$args,$ac_parser)=@_;
    my $storage=$ac_parser->{-storage};
    my $vars=$ac_parser->{-vardata};
    my $lang_flag=$LANG_FLAGS{$name};
    my $lib_flag=$LIB_FLAGS{$name};
    my $pkg_flag=$PKG_FLAGS{$name};
    my $bin_flag=$BIN_FLAGS{$name};
    my $header_flag=$HEADER_FLAGS{$name};
    $storage->{lang}->{$lang_flag}=1 if $lang_flag;
    push @{$storage->{libsreq}}, split(';',$lib_flag) if $lib_flag;
    push @{$storage->{pkgsreq}}, split(';',$pkg_flag) if $pkg_flag;
    push @{$storage->{pathreq}}, split(';',$bin_flag) if $bin_flag;
    push @{$storage->{hdrsreq}}, &__add_h(split(';',$header_flag)) if $header_flag;
    warn "SourceAnalyzer: unknown macros $name. Please. report.\n" if $name=~/^A[CM]_PATH_(?!PROG)/ and not ($lang_flag or $lib_flag or $header_flag or $pkg_flag or $IGNORE_FLAGS{$name});

    if ($name eq 'AC_LANG') {
	my $lang=__strip($args->[0]);
	if ($lang eq 'C') {
	    $storage->{lang}->{'c'}=1;
	} elsif ($lang eq 'C++') {
	    $storage->{lang}->{'cxx'}=1;
	} elsif ($lang =~ '^Fortran') {
	    $storage->{lang}->{'fortran'}=1;
	} else {
	    warn "SourceAnalyzer bug: unknown language in AC_LANG($lang). Please, report.\n";
	}
    } elsif ($name eq 'AC_CHECK_LIB') {
	&__push($storage->{libsreq}, &__subst_vars($args->[0],$vars));
    } elsif ($name eq 'AC_SEARCH_LIBS') {
	&__push($storage->{libsreq}, [&__split(&__subst_vars($args->[1],$vars))]);
    } elsif ($name eq 'AC_CHECK_HEADER') {
	&__push($storage->{hdrsreq}, &__add_h(&__subst_vars($args->[0], $vars)));
    } elsif ($name eq 'AC_CHECK_HEADERS') {
	&__push($storage->{hdrsreq}, &__add_h(&__split(&__subst_vars($args->[0],$vars))));
    } elsif ($name eq 'AC_PATH_GENERIC') {
	&__push($storage->{pathreq}, map {$_.'-config'} &__subst_vars($args->[1],$vars));
    } elsif ($name eq 'AC_PATH_PROG') {
	&__push($storage->{pathreq}, &__subst_vars($args->[1],$vars));
    } elsif ($name eq 'AC_CHECK_PROG') {
	&__push($storage->{pathreq}, &__subst_vars($args->[1],$vars));
    } elsif ($name eq 'AC_CHECK_PROGS') {
	&__push($storage->{pathreq}, [&__split(&__subst_vars($args->[1],$vars))]);
    } elsif ($name eq 'PKG_CHECK_MODULES') {
	my $pkgline=&__subst_vars($args->[1],$vars);
# note 3 arg
#PKG_CHECK_MODULES([ENCHANT], [enchant], [AC_DEFINE(HAVE_LIBENCHANT, , Define if we use enchant)], [])
# TODO: multiline (see assogiate)
#PKG_CHECK_MODULES([ASSOGIATE], [                                                
#        glib-2.0 >= 2.8.0
#        glibmm-2.4 >= 2.8.0
#        gtkmm-2.4 >= 2.6.0
#        libxml++-2.6
#        gnome-vfsmm-2.6
#])                                                                              
# TODO: variable (see xqf); note: it is inside if; so it still might be better to ignore
#        pkg_modules="gtk+-2.0 >= 2.0.0 gdk-pixbuf-xlib-2.0"
#        PKG_CHECK_MODULES(PACKAGE, [$pkg_modules])


	my @raw_pkgconf;
	foreach my $item (grep {$_} split (/\s+/, $pkgline)) {
	    if ($item=~/^.+\$([\w_][\w_\d]*)/) {
		if (not $vars->{$1}) {
		    push @raw_pkgconf, $item;
		} else {
		    #print STDERR "var: $1 - ",join(':',@{$vars->{$1}}),"\n";
		    foreach (@{$vars->{$1}}) {
			my $expanded_item=$item;
			$expanded_item=~ s/^.+\K\$([\w_][\w_\d]*)/$_/;
			push @raw_pkgconf, $expanded_item;
		    }
		}
	    } else {
		push @raw_pkgconf, $item;
	    }
	}
	while (@raw_pkgconf>0) {
	    my $in=shift @raw_pkgconf;
	    next if $in =~ /^\$/;
	    $in =~s/^['"]//;
	    $in =~s/['"]$//;
	    if ($in =~ /^[<>=!]/) {
		shift @raw_pkgconf;
		next;
	    }
	    push @{$storage->{pkgconfreq}}, $in;
	}
    }
}

sub __subst_vars {
    my ($in,$vars)=@_;
    my $line=&__strip($in);
    $line=~s/(?:^|\s)\K\$([\w_][\w_\d]*)/$vars->{$1} ? join(' ',@{$vars->{$1}}) : $1/eg;
    return $line;
}

sub __push {
    my ($ref,@in)=@_;
    map {Carp::cluck "undefined value" unless defined $_} @_;
    push @$ref, @in;
}

sub __strip {
    local $_=$_[0];
    s/\\\n//gs;
    s/^\s*//;
    s/\s*$//;
    return $_;
}

sub __split {
    my ($line)=@_;
    return grep {defined($_) and $_ ne ''} split (/\s+/, __strip($line));
}

sub __add_h {
    # not used...
    #return map {/\.h$/ ? $_ : $_.'.h'} @_;
    return @_;
}

sub __add_file_to_flag {
    my ($file,$hash)=@_;
    return unless -e $file;
    open my $fh, "<", $file or die $!;
    local $_;
    while (<$fh>) {
	chomp;
	s/^\s*//;
	s/\s*$//;
	$hash->{$_}=1;
    }
    close $fh;
}

sub _automake_var_handler {
    my ($name,$args,$am_parser)=@_;
    my $storage=$am_parser->{-storage};
    if ($am_parser->{-verbose}>3) {
	print "_automake_var_handler: $name=$args\n";
    }
    if ($name=~/_(?:LIBS|LDFLAGS|LDADD|LIBADD)$/) {
	foreach my $word (shellwords($args)) {
	    &__push($storage->{libsreq}, $word) if $word=~s/^-l(?=\w)//;
	}
    }
}

sub _qmake_var_handler {
    my ($name,$args,$parser)=@_;
    my $storage=$parser->{-storage};
    print "_qmake_var_handler: $name=$args\n" if $parser->{-verbose}>3;
    if ($name eq 'LIBS') {
     	foreach my $word (shellwords($args)) {
     	    #&__push($storage->{libsreq}, $word) if $word=~s/^-l(?=\w)//;
     	    print STDERR "qmake project: TODO: need lib${word}.so\n" if $word=~s/^-l(?=\w)// and $parser->{-verbose};
     	}
    } elsif ($name eq 'PKGCONFIG') {
     	foreach my $word (shellwords($args)) {
     	    &__push($storage->{libsreq}, $word) if $word=~/^\w/;
     	}
    } elsif ($name eq 'TEMPLATE') {
	# if 
        #TEMPLATE = lib
	# then put target to -ignore section (internal lib)?
	# or just use flag 'ban self'
    } elsif ($name eq 'TARGET') {
	# TODO
    }
}

package __::SourceAnalyzer::Collect;

sub new {
    my $class = shift;
    my $self = {
	libsreq=>[],
	hdrsreq=>[],
	pathreq=>[],
	pkgsreq=>[],
	pkgconfreq=>[],
	lang=>{},
	@_
    };
    bless $self, $class;
}

package SourceAnalyzer::Language;

use strict;
use warnings;
our $VERSION=0.001;

sub parse_file {
    my ($self,$file)=@_;
    my $content = &__load_file($file);
    return $self->parse_string($content);
}

sub __load_file {
    my ($file)=@_;
    open my $fh, "<", $file or die $!;
    local $/; # enable localized slurp mode
    my $content = <$fh>;
    close $fh;
    return $content;
}

package SourceAnalyzer::Language::Autoconf;

use strict;
use warnings;
our $VERSION=0.001;
our @ISA=qw/SourceAnalyzer::Language/;

# for copypaste from Text/ParseWords.pm; seems to be support for perlish syntax not useful for us.
my $PERL_SINGLE_QUOTE;
our $verbose=0;

sub new {
    my $class = shift;
    my $self = {
	-vars => {},
	-verbose => 0,
	-varhook => \&_var_handler,
	@_
    };
    bless $self, $class;
}

sub parse_string {
    my ($self,$data)=@_;
    $self->{-vardata}={};
    while (length($data)) {
	&_read_any($self,\$data);
    }
}

# C++ features
#CXXAC_TYPE_BOOL
#CXXAC_BAD_ALLOC
#CXXAC_HAVE_EXPLICIT
#CXXAC_HAVE_MUTABLE
#CXXAC_HAVE_ARRAY_NEW
#CXXAC_BAD_DELETE
#CXXAC_DEFAULT_EXCEPTION
#CXXAC_HAVE_TYPENAME
#CXXAC_NO_GUIDE_DECL
#CXXAC_HEADER_ALL

my %KNOWN_MACRO=map {$_=>1} qw!
DAP_CHECK_ZLIB
GNOME_COMMON_INIT
GNOME_COMPILE_WARNINGS
GNOME_DEBUG_CHECK
GNOME_DOC_INIT
PKG_CHECK_MODULES
PKG_PROG_PKG_CONFIG
GTK_DOC_CHECK
AST_IMLIB2_SUPPORT
!;

sub _read_any {
    my ($self,$ref)=@_;
    my $read=&_read_non_token_non_end_or_comments($ref);
    return $read if $read ne '';
    $read=$1 if $$ref=~s/^([,\)])//;
    return $read if $read ne '';
    my $token=&_read_token($ref);
    warn "Oops! no token" if $token eq '';
    $read.=$token;
    if ($token eq 'dnl') {
	$$ref=~s/^([^\n]*\n)//s;
	return $read.$1;
    } elsif ($KNOWN_MACRO{$token} or $token=~/^(?:AC_|AM_|CXXAC_)/) {
	my $args=[];
	$read.=&_read_args($self,$ref,$args) if $$ref=~/^\(/;
	$self->add_macro($token,$args);
    } elsif ($$ref=~s/^=//) {
	my ($unquoted, $raw)=&_read_shellword('\s+', $ref);
	$self->add_var($token,$unquoted);
	$read .= '='.$raw;
    }
    return $read;
}

sub add_macro {
    my ($self,$name,$args)=@_;
    print STDERR "MACRO: $name ARGS:",join(',',@$args),"\n" if $verbose>1;
    my $hook=$self->{-macrohook};
    &$hook($name,$args,$self) if $hook;
}

sub add_var {
    my ($self,$name,$val)=@_;
    print STDERR "VAR: add $name=$val\n" if $verbose>1;
    my $hook=$self->{-varhook};
    &$hook($name,$val,$self) if $hook;
    return $val;
}

sub _var_handler {
    my ($name,$val,$self)=@_;
    my $VAR=$self->{-vardata};
    my $ref=$VAR->{$name};
    unless ($ref) {
	$ref=[];
	$VAR->{$name}=$ref;
    }
    push @$ref, $val;
}

sub _read_token {
    my ($ref)=@_;
    if ($$ref=~s/^([\w_][\w_\d]*)//s) {
	print STDERR "read token: $1\n" if $verbose>2;
	return $1;
    } else {
	return '' if $$ref eq '';
	die "Oops! read_token - no token :( ->".substr($$ref,0,1);
    }
}

sub _read_non_token_non_end_or_comments {
    my ($ref)=@_;
    my $read='';
    while (length($$ref) and $$ref !~/^[\w_\),]/) {
	if ($$ref=~s/^(\s+)//s) {
	    $read.=$1;
	} elsif ($$ref=~s/^(#[^\n]*)\n//s) {
	    print STDERR "drop comment: $1\n" if $verbose>2;
	    $read.=$1;
	} else {
	    $$ref=~s/^(.)//s;
	    $read.=$1;
	}
    }
    return $read;
}

sub _read_args {
    my ($self,$ref,$arg)=@_;
    my $read='';
    my $cur_arg='';
    if ($$ref=~s/^(\()//s) {
	$read.=$1;
    }
    while (length($$ref) and $$ref !~/^\)/) {
	if ($$ref=~s/^(\s+)//s) {
	    $read.=$1;
	    $cur_arg.=$1;
	# TODO: properly read m4 quoted string
	# TODO: load shell var assignments in m4 quoted string
	} elsif ($$ref=~s/^(\[([^]]*)\])//s) {
	    $read.=$1;
	    $cur_arg.=$2;
	} elsif ($$ref=~s/^(,)//s) {
	    $read.=$1;
	    push @$arg, $cur_arg;
	    $cur_arg='';
	} else {
	    $cur_arg.=&_read_any($self,$ref);
	}
    }
    if ($$ref=~s/^(\))//s) {
	$read.=$1;
    }
    push @$arg, $cur_arg;
    return $read;
}

# regexp is copypasted from Text/ParseWords.pm
sub _read_shellword {
    my($delimiter, $lineref) = @_;
    my($word);

    no warnings 'uninitialized';	# we will be testing undef strings

    return '' if not length($$lineref);
        # This pattern is optimised to be stack conservative on older perls.
        # Do not refactor without being careful and testing it on very long strings.
        # See Perl bug #42980 for an example of a stack busting input.
    $$lineref =~ s/^
                    (?: 
                        # double quoted string
                        (")                             # $quote
                        ((?>[^\\"]*(?:\\.[^\\"]*)*))"   # $quoted 
		    |	# --OR--
                        # singe quoted string
                        (')                             # $quote
                        ((?>[^\\']*(?:\\.[^\\']*)*))'   # $quoted
                    |   # --OR--
                        # unquoted string
		        (                               # $unquoted 
                            (?:\\.|[^\\"'])*?           
                        )		
                        # followed by
		        (                               # $delim
                            \Z(?!\n)                    # EOL
                        |   # --OR--
                            (?-x:$delimiter)            # delimiter
                        |   # --OR--                    
                            (?!^)(?=["'])               # a quote
                        )  
		    )//xs or return;		# extended layout                  
        my ($quote, $quoted, $unquoted, $delim) = (($1 ? ($1,$2) : ($3,$4)), $5, $6);


	return '' unless( defined($quote) || length($unquoted) || length($delim));

    my $raw = "$quote$quoted$quote$delim";

    $unquoted =~ s/\\(.)/$1/sg;
    if (defined $quote) {
	$quoted =~ s/\\(.)/$1/sg if ($quote eq '"');
	$quoted =~ s/\\([\\'])/$1/g if ( $PERL_SINGLE_QUOTE && $quote eq "'");
    }
    my $processed = defined $quote ? $quoted : $unquoted;
    return ($processed, $raw);
}

package SourceAnalyzer::Language::AutoMake;

use strict;
use warnings;

our @ISA=qw/SourceAnalyzer::Language/;
our $verbose=0;
our $VERSION=0.001;

sub new {
    my $class = shift;
    my $self = {
	-verbose => 0,
	@_
    };
    bless $self, $class;
}

sub parse_string {
    my ($self,$data)=@_;
    $data=~s/\\\n//gs;
    foreach my $line (split("\n",$data)) {
	&_parse_line($self,$line);
    }
}

sub _parse_line {
    my ($self,$line)=@_;
    $line=~s/^\s*//;
    $line=~s/\s*$//;
    if ($line=~/([\w_][\w\d_]*)\s*\+?=\s*(.*)$/) {
	$self->add_var($1,$2);
    }
}

sub add_var {
    my ($self,$name,$val)=@_;
    print STDERR "VAR: add $name=$val\n" if $verbose>1;
    my $hook=$self->{-varhook};
    &$hook($name,$val,$self) if $hook;
    return $val;
}

package SourceAnalyzer::Language::QMake;

use strict;
use warnings;

our @ISA=qw/SourceAnalyzer::Language/;
our $verbose=0;
our $VERSION=0.001;

sub new {
    my $class = shift;
    my $self = {
	-verbose => 0,
	@_
    };
    bless $self, $class;
}

sub parse_string {
    my ($self,$data)=@_;
    $data=~s/\\\n//gs;
    foreach my $line (split("\n",$data)) {
	&_parse_line($self,$line);
    }
}

sub _parse_line {
    my ($self,$line)=@_;
    $line=~s/^\s*//;
    $line=~s/\s*$//;
    if ($line=~/([\w_][\w\d_]*)\s*\+?=\s*(.*)$/) {
	$self->add_var($1,$2);
    }
}

sub add_var {
    my ($self,$name,$val)=@_;
    print STDERR "VAR: add $name=$val\n" if $verbose>1;
    my $hook=$self->{-varhook};
    &$hook($name,$val,$self) if $hook;
    return $val;
}

1;
